/*
 * Copyright (C) 2002 Sven Schaepe <schaepe@rz.tu-ilmenau.de>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *
 * 2004-02-11: Ondrej Bojar <obo@cuni.cz> added more colors on the bars
 *               and failed to fix a bug that damages the background source
 *               bitmap if a bar gets full. The workaround was to shift
 *               the source bars in the bitmap more to the right so that they
 *               do not get damaged.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <linux/version.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>

#include "../wmgeneral/wmgeneral.h"
#include "../wmgeneral/misc.h"

#include "wmsm-master.xpm"

#define v_WMSM "0.2.1"

/* scale */
#define xpm_scale_width		35
#define xpm_scale_height	7
/* type 1 */
#define xpm_scale_1_f_x		79
#define xpm_scale_1_f_y		5
#define xpm_scale_1_b_x		79
#define xpm_scale_1_b_y		13
/* type 2 */
#define xpm_scale_2_f_x		79
#define xpm_scale_2_f_y		21
#define xpm_scale_2_b_x		79
#define xpm_scale_2_b_y		29
/* red, type 1 and 2 */
#define xpm_scale_red_1_y	    37
#define xpm_scale_red_2_y	    45
#define xpm_scale_yellow_1_y    53
#define xpm_scale_yellow_2_y    61
#define xpm_scale_green_1_y	    69
#define xpm_scale_green_2_y	    77

/* digit */
#define digit_width		6
#define digit_height		7
/* green */
#define digit_1_x		5
#define digit_1_y		66
/* blue */
#define digit_2_x		5
#define digit_2_y		76

#define cpu_s_x			24
#define cpu_s_y			5
#define mem_s_x			24
#define mem_s_y			14
#define swap_s_x		24
#define swap_s_y		23
#define ior_s_x			24
#define ior_s_y			32
#define iow_s_x			24
#define iow_s_y			41

typedef struct {
    int cpu_idle;
    int cpu_use;
    int user;
    int system;
    int nice;
    long long r_diff_max;
    unsigned long long r_io;
    long long w_diff_max;
    unsigned long long w_io;
    int day;
    int hour;
    int min;
    int sec;
} LAST;
LAST last;

char wmsm_mask_bits[64*64];
int wmsm_mask_width = 64;
int wmsm_mask_height = 64;
int scale_width;
int scale_height;
int scale_f_x;
int scale_f_y;
int scale_b_x;
int scale_b_y;
int scale_red_y;
int scale_yellow_y;
int scale_green_y;

static void draw_digit(int value,int dx,int dy)
{
    if(value>99) {
	copyXPMArea(9*digit_width+digit_2_x,digit_2_y,digit_width,digit_height,dx,dy);
	copyXPMArea(9*digit_width+digit_2_x,digit_2_y,digit_width,digit_height,dx+digit_width,dy);
    } else {
	copyXPMArea((value/10)*digit_width+digit_1_x,digit_1_y,digit_width,digit_height,dx,dy);
	copyXPMArea((value%10)*digit_width+digit_1_x,digit_1_y,digit_width,digit_height,dx+digit_width,dy);
    }
    RedrawWindowXY(0,0);
}

static void draw_uptime(int day,int hour,int min,int sec)
{
    if(sec!=last.sec) {
	draw_digit(sec,48,51);
	last.sec=sec;
	if(min!=last.min) {
	    draw_digit(min,34,51);
	    last.min=min;
	    if(hour!=last.hour) {
		draw_digit(hour,20,51);
		last.hour=hour;
		if(day!=last.day) {
		    draw_digit(day,5,51);
		    last.day=day;
		}
	    }
	}
    } 
}

static void draw_scale(int dx,int dy,
  int src1y, int width1, int src2y, int width2, int src3y, int width3)
{
    int width=0;
    /* range checks */
    if (width1<0) width1 = 0;
    if (width2<0) width2 = 0;
    if (width3<0) width3 = 0;
    if ((width = width1+width2+width3) > scale_width) {
      width1 = (int)(width1*scale_width/width);
      width2 = (int)(width2*scale_width/width);
      width3 = (int)(width3*scale_width/width);
    }
    width = 0;
    if (width1)
      copyXPMArea(scale_f_x,src1y,width1,scale_height,dx,dy);
    width+=width1;
    if (width2)
      copyXPMArea(scale_f_x+width,src2y,width2+width,scale_height,dx+width,dy);
    width+=width2;
    if (width3)
      copyXPMArea(scale_f_x+width,src3y,width3+width,scale_height,dx+width,dy);
    width+=width3;
    /* blank the rest */
    if(width<scale_width) {
	copyXPMArea(scale_b_x+width,scale_b_y,scale_width-width,scale_height,dx+width,dy);
    }
    RedrawWindowXY(0,0);
}

static int getDiskValue24(unsigned long long *r_io,unsigned long long *w_io)
{
    FILE *file;
    char line[256];
    char tag[32];
    int major=0,disk=0,all_io=0,r_blk=0,w_blk=0;
    if((file=fopen("/proc/stat","r"))==NULL) {
	return -1;
    }
    while(fgets(line,sizeof(line),file)!=NULL) {
	sscanf(line,"%s",tag);
	if(strcmp(tag,"disk_io:")==0) {
	    sscanf(line,"%s (%d,%d):(%d,%llu,%d,%llu,%d)",tag,&major,&disk,&all_io
						    ,r_io,&r_blk,w_io,&w_blk);
	    fclose(file);
	    return 0;
	}
    }
    fclose(file);
    return -1;
}

static int getDiskValue26(unsigned long long *r_io,unsigned long long *w_io)
{
    FILE *file;
    char line[256];
    if((file=fopen("/proc/diskstats","r"))==NULL) {
	return -1;
    }
    while(fgets(line,sizeof(line),file)!=NULL) {
	int major,minor;
	char diskName[40];
	unsigned int reads,readMerges,readTicks;
	unsigned int writes,writeMerges,writeTicks;
	unsigned int inFlight,ioTicks,timeInQueue;
	unsigned long long readSectors,writeSectors;
	if(sscanf(line,"%4d %4d %s %u %u %llu %u %u %u %llu %u %u %u %u",
	    &major,&minor,diskName,&reads,&readMerges,&readSectors,&readTicks,
	    &writes,&writeMerges,&writeSectors,&writeTicks,&inFlight,&ioTicks,
	    &timeInQueue)==14) {
	    *r_io=readSectors;
	    *w_io=writeSectors;
	    fclose(file);
	    return 0;
	}
    }
    fclose(file);
    return -1;
}

static void get_values(int init,int kernelVersion)
{
    FILE *file;
    char line[256];
    char tag[32];
    int user,nice,system,idle;
    unsigned long long r_io=0,w_io=0;
    int mem_total=0,mem_free=0,mem_cached=0,mem_buffers=0;
    int swap_total=0,swap_free=0;
    float uptime=0.0;

    file=fopen("/proc/stat","r");
    while(fgets(line,sizeof(line),file)!=NULL) {
	sscanf(line,"%s",tag);
	if(strcmp(tag,"cpu")==0) {
	    sscanf(line,"%s %d %d %d %d",tag,&user,&nice,&system,&idle);
	    break;
	}
    }
    fclose(file);
    if(kernelVersion<KERNEL_VERSION(2,6,0)) {
	getDiskValue24(&r_io,&w_io);
    } else {
	getDiskValue26(&r_io,&w_io);
    }
    file=fopen("/proc/meminfo","r");
    while(fgets(line,sizeof(line),file)!=NULL) {
	sscanf(line,"%s",tag);
	if(strcmp(tag,"MemTotal:")==0) {
	    sscanf(line,"%s %d",tag,&mem_total);
	    continue;
	}
	if(strcmp(tag,"MemFree:")==0) {
	    sscanf(line,"%s %d",tag,&mem_free);
	    continue;
	}
	if(strcmp(tag,"Cached:")==0) {
	    sscanf(line,"%s %d",tag,&mem_cached);
	    continue;
	}
	if(strcmp(tag,"Buffers:")==0) {
	    sscanf(line,"%s %d",tag,&mem_buffers);
	    continue;
	}
	if(strcmp(tag,"SwapTotal:")==0) {
	    sscanf(line,"%s %d",tag,&swap_total);
	    continue;
	}
	if(strcmp(tag,"SwapFree:")==0) {
	    sscanf(line,"%s %d",tag,&swap_free);
	    break;
	}
    }
    fclose(file);

    file=fopen("/proc/uptime","r");
    fscanf(file,"%f",&uptime);
    fclose(file);

    if(init) {
	last.cpu_idle=idle;
	last.cpu_use=user+nice+system;
	last.user=user;
	last.nice=nice;
	last.system=system;
	last.r_diff_max=1;
	last.r_io=r_io;
	last.w_diff_max=1;
	last.w_io=w_io;
    } else {
	long long diff;
	int width;
	int mem_used_width;
	int mem_buffers_width;
	int mem_cached_width;
    long long cpu_total;
    int user_width;
    int system_width;
    int nice_width;

	/* cpu */
    cpu_total = idle-last.cpu_idle + user+system+nice -last.cpu_use +1;
    user_width = (int)(((float)user - (float)last.user)*(float)scale_width/(float)cpu_total);
    system_width = (int)(((float)system - (float)last.system)*(float)scale_width/(float)cpu_total);
    nice_width = (int)(((float)nice - (float)last.nice)*(float)scale_width/(float)cpu_total);
	last.cpu_idle=idle;
	last.cpu_use=user+nice+system;
	last.user=user;
	last.nice=nice;
	last.system=system;
	draw_scale(cpu_s_x,cpu_s_y,scale_red_y,system_width,scale_green_y,user_width,scale_yellow_y,nice_width);

	/* r_io */
	diff=r_io-last.r_io;
	if(diff>last.r_diff_max) {
	    last.r_diff_max=diff;
	}
	width=abs(diff)*scale_width/last.r_diff_max;
	last.r_io=r_io;
	draw_scale(ior_s_x,ior_s_y,scale_f_y,width,0,0,0,0);
	
	/* w_io */
	diff=w_io-last.w_io;
	if(diff>last.w_diff_max) {
	    last.w_diff_max=diff;
	}
	width=abs(diff)*scale_width/last.w_diff_max;
	last.w_io=w_io;
	draw_scale(iow_s_x,iow_s_y,scale_red_y,width,0,0,0,0);
	
	/* mem */
	if(!mem_total) {
	    mem_used_width=0;
	    mem_cached_width=0;
	    mem_buffers_width=0;
	} else {
        mem_used_width=(mem_total-mem_free-mem_cached-mem_buffers)*scale_width/mem_total;
        mem_cached_width=mem_cached*scale_width/mem_total;
        mem_buffers_width=mem_buffers*scale_width/mem_total;
	}
	draw_scale(mem_s_x,mem_s_y,scale_red_y,mem_used_width,scale_yellow_y,mem_buffers_width,scale_f_y,mem_cached_width);
	
	/* swap */
	if(!swap_total) {
	    width=0;
	} else {
	    width=(swap_total-swap_free)*scale_width/swap_total;
	}
	draw_scale(swap_s_x,swap_s_y,scale_red_y,width,0,0,0,0);

	/* uptime */
	{
	    int day,hour,min,sec;

	    day=(int)uptime/(24*3600);
	    uptime-=day*24*3600;
	    hour=(int)uptime/3600;
	    uptime-=hour*3600;
	    min=(int)uptime/60;
	    uptime-=min*60;
	    sec=uptime;
	    draw_uptime(day,hour,min,sec);
	}
    }
 }

static void usage(char *progname)
{
    fprintf(stdout,"%s-%s "
		    "WindowMaker System Monitor\n"
		    "\t-t <num>	scale type num={1,2,3,4}\n"
		    "\t-h		this\n"
		    ,progname,v_WMSM);
}

int main(int argc,char *argv[])
{
    FILE *file;
    char line[256];
    char xx[256];
    int revision=0,release=0,patchlevel=0;
    XEvent event;
    int opt;

    scale_width=xpm_scale_width;
    scale_height=xpm_scale_height;
    scale_f_x=xpm_scale_2_f_x;
    scale_f_y=xpm_scale_2_f_y;
    scale_b_x=xpm_scale_1_b_x;
    scale_b_y=xpm_scale_1_b_y;
    scale_red_y=xpm_scale_red_2_y;
    scale_yellow_y=xpm_scale_yellow_2_y;
    scale_green_y=xpm_scale_green_2_y;

    while((opt=getopt(argc,argv,"hmt:"))!=-1) {
	switch(opt) {
		break;
	    case 't':
		switch(atoi(optarg)) {
		    case 4:
			scale_f_x=xpm_scale_2_f_x;
			scale_f_y=xpm_scale_2_f_y;
			scale_b_x=xpm_scale_2_b_x;
			scale_b_y=xpm_scale_2_b_y;
            scale_red_y=xpm_scale_red_2_y;
            scale_yellow_y=xpm_scale_yellow_2_y;
            scale_green_y=xpm_scale_green_2_y;
			break;
		    case 3:
			scale_f_x=xpm_scale_1_f_x;
			scale_f_y=xpm_scale_1_f_y;
			scale_b_x=xpm_scale_1_b_x;
			scale_b_y=xpm_scale_1_b_y;
            scale_red_y=xpm_scale_red_1_y;
            scale_yellow_y=xpm_scale_yellow_1_y;
            scale_green_y=xpm_scale_green_1_y;
			break;
		    case 2:
			scale_f_x=xpm_scale_1_f_x;
			scale_f_y=xpm_scale_1_f_y;
			scale_b_x=xpm_scale_2_b_x;
			scale_b_y=xpm_scale_2_b_y;
            scale_red_y=xpm_scale_red_1_y;
            scale_yellow_y=xpm_scale_yellow_1_y;
            scale_green_y=xpm_scale_green_1_y;
			break;
		    case 1:
		    default:
			scale_f_x=xpm_scale_2_f_x;
			scale_f_y=xpm_scale_2_f_y;
			scale_b_x=xpm_scale_1_b_x;
			scale_b_y=xpm_scale_1_b_y;
            scale_red_y=xpm_scale_red_2_y;
            scale_yellow_y=xpm_scale_yellow_2_y;
            scale_green_y=xpm_scale_green_2_y;
		}
		break;
	    case 'h':
	    default:
		usage(argv[0]);
		exit(0);
	}
    }

    file=fopen("/proc/version","r");
    fgets(line,sizeof(line),file);
    sscanf(line,"%s %s %d.%d.%d%s",xx,xx,&revision,&release,&patchlevel,xx);
    fclose(file);

    createXBMfromXPM(wmsm_mask_bits,wmsm_master_xpm,wmsm_mask_width,wmsm_mask_height);
    openXwindow(argc,argv,wmsm_master_xpm,wmsm_mask_bits,wmsm_mask_width,wmsm_mask_height);
    RedrawWindowXY(0,0);

    get_values(1,(revision<<16)+(release<<8)+patchlevel);
    while(1) {
	while (XPending(display)) {
	    XNextEvent(display, &event);
	    switch (event.type) {
		case Expose:
		    RedrawWindowXY(0,0);
		    break;
		case DestroyNotify:
		    XCloseDisplay(display);
		    exit(0);
	    }
	}
	get_values(0,(revision<<16)+(release<<8)+patchlevel);
	usleep(250000L);
    }
}
